/*
Copyright (c) 1999 Rafal Wojtczuk <nergal@icm.edu.pl>. All rights reserved.
See the file COPYING for license details.
*/

#include <pcap.h>
#include <malloc.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <syslog.h>
#include "util.h"
#include "nids.h"
#include "ip_fragment.h"
#include "scan.h"

extern char *inet_ntoa (long);

char errbuf[PCAP_ERRBUF_SIZE];

static struct proc_node *ip_frag_procs;
static struct proc_node *ip_procs;
struct proc_node *tcp_procs;
extern int tcp_init (int);
extern int ip_frag_init (int);
extern void scan_init ();
extern void process_tcp (u_char *, int);
extern int ip_defrag_stub (char *, struct iphdr **);
extern int ip_options_compile (char *);
static int linkoffset = 0;

char *nids_warnings[] =
{
  "Murphy - you never should see this message !",
  "Oversized IP packet",
  "Invalid IP fragment list: fragment over size",
  "Overlapping IP fragments",
  "Invalid IP header",
  "Source routed IP frame",
  "Max number of TCP streams reached",
  "Invalid TCP header",
  "Too much data in TCP receive queue",
  "Invalid TCP flags"
};

static void nids_syslog (int, int, struct iphdr *, void *);

struct nids_prm nids_params =
{1024, 256, 0, 168, -1, nids_syslog, LOG_ALERT, 256, 3000, 10};

static void
nids_syslog (int type, int errnum, struct iphdr *iph, void *data)
{
  char saddr[20], daddr[20];
  char buf[1024];
  struct host *this_host;
  unsigned char flagsand = 255, flagsor = 0;
  int i;
  switch (type)
    {
    case NIDS_WARN_IP:
      if (errnum != NIDS_WARN_IP_HDR)
	{
	  strcpy (saddr, inet_ntoa (iph->saddr));
	  strcpy (daddr, inet_ntoa (iph->daddr));
	  syslog (nids_params.syslog_level,
	   "%s, packet (apparently) from %s to %s\n", nids_warnings[errnum],
		  saddr, daddr);
	}
      else
	syslog (nids_params.syslog_level, "%s\n", nids_warnings[errnum]);
      break;
    case NIDS_WARN_TCP:
      strcpy (saddr, inet_ntoa (iph->saddr));
      strcpy (daddr, inet_ntoa (iph->daddr));
      if (errnum != NIDS_WARN_TCP_HDR)
	syslog (nids_params.syslog_level,
		"%s,from %s:%hi to  %s:%hi\n", nids_warnings[errnum], saddr,
		ntohs (((struct tcphdr *) data)->source), daddr,
		ntohs (((struct tcphdr *) data)->dest));
      else
	syslog (nids_params.syslog_level, "%s,from %s to %s\n",
		nids_warnings[errnum], saddr, daddr);
      break;
    case NIDS_WARN_SCAN:
      this_host = (struct host *) data;
      sprintf (buf, "Scan from %s. Scanned ports: ", inet_ntoa (this_host->addr));
      for (i = 0; i < this_host->n_packets; i++)
	{
	  strcat (buf, inet_ntoa (this_host->packets[i].addr));
	  sprintf (buf + strlen (buf), ":%hi,", this_host->packets[i].port);
	  flagsand &= this_host->packets[i].flags;
	  flagsor |= this_host->packets[i].flags;
	}
      if (flagsand == flagsor)
	{
	  i = flagsand;
	  switch (flagsand)
	    {
	    case 2:
	      strcat (buf, "scan type: SYN");
	      break;
	    case 0:
	      strcat (buf, "scan type: NULL");
	      break;
	    case 1:
	      strcat (buf, "scan type: FIN");
	      break;
	    default:
	      sprintf (buf + strlen (buf), "flags=0x%x", i);
	    }
	}
      else
	strcat (buf, "various flags");
      syslog (nids_params.syslog_level, "%s", buf);
      break;


    default:
      syslog (nids_params.syslog_level, "Unknown warning number ?\n");
    }
}





static void
pcap_hand (u_char * par, struct pcap_pkthdr *hdr, u_char * data)
{
  struct proc_node *i;
  if (!strncmp(nids_params.device,"eth",3) && 
      (data[12]!=8 || data[13]!=0))
  return; //not an IP ethernet frame
  for (i = ip_frag_procs; i; i = i->next)
    (i->item) (data + linkoffset, hdr->len - linkoffset);
}

static void
gen_ip_frag_proc (u_char * data, int len)
{
  struct proc_node *i;
  struct iphdr *iph = (struct iphdr *) data;
  int need_free=0;
  int skblen;
  if (len < sizeof (struct iphdr) || iph->ihl < 5 || iph->version != 4 ||
      ip_fast_csum ((unsigned char *) iph, iph->ihl) != 0
      || len < ntohs (iph->tot_len) ||
      ntohs (iph->tot_len) < iph->ihl << 2)
    {
      nids_params.syslog (NIDS_WARN_IP, NIDS_WARN_IP_HDR, iph, 0);
      return;
    }
  if (iph->ihl > 5 && ip_options_compile (data))
    {
      nids_params.syslog (NIDS_WARN_IP, NIDS_WARN_IP_SRR, iph, 0);
      return;
    }
  switch (ip_defrag_stub (data, &iph))
    {
    case IPF_ISF:
      return;
    case IPF_NOTF:
      need_free = 0;
      iph = (struct iphdr *) data;
      break;
    case IPF_NEW:
      need_free = 1;
      break;
    default:;
    }
  skblen = ntohs (iph->tot_len) + 16;
  if (!need_free)
    skblen += nids_params.dev_addon;
  skblen = (skblen + 15) & ~15;
  skblen += nids_params.sk_buff_size;

  for (i = ip_procs; i; i = i->next)
    (i->item) (iph, skblen);
  if (need_free)
    free (iph);
}

static void 
process_udp (char *x)
{
}

static void
gen_ip_proc (u_char * data, int skblen)
{
  switch (((struct iphdr *) data)->protocol)
    {
    case IPPROTO_TCP:
      process_tcp (data, skblen);
      break;
    case IPPROTO_UDP:
      process_udp (data);
      break;
    default:;
    }
}


static void
init_procs ()
{
  ip_frag_procs = mknew (struct proc_node);
  ip_frag_procs->item = gen_ip_frag_proc;
  ip_frag_procs->next = 0;
  ip_procs = mknew (struct proc_node);
  ip_procs->item = gen_ip_proc;
  ip_procs->next = 0;
  tcp_procs = 0;
}

void
nids_register_ip (void (*x))
{
  struct proc_node *ipp = mknew (struct proc_node);
  ipp->item = x;
  ipp->next = ip_procs;
  ip_procs = ipp;
}

void
nids_register_ip_frag (void (*x))
{
  struct proc_node *ipp = mknew (struct proc_node);
  ipp->item = x;
  ipp->next = ip_frag_procs;
  ip_procs = ipp;
}

static pcap_t *desc;

void
nids_init ()
{
  char *device = nids_params.device;
  if (!device)
    device = pcap_lookupdev (errbuf);
  if (!device)
    {
      fprintf (stderr, "%s", errbuf);
      exit (1);
    }
  nids_params.device=device;  
  if (!strncmp (device, "eth", 3))
    linkoffset = 14;
  else
    linkoffset = 0;
  if (nids_params.dev_addon == -1)
    if (!strncmp (device, "ppp", 3))
      nids_params.dev_addon = 0;
    else
      nids_params.dev_addon = 16;
  desc = pcap_open_live (device, 16384, 1, 0, errbuf);
//desc=pcap_open_offline("dupadupa",errbuf);
  if (!desc)
    {
      fprintf (stderr, "%s", errbuf);
      exit (1);
    }
  if (nids_params.syslog == nids_syslog)
    openlog ("libnids", 0, LOG_LOCAL0);
  init_procs ();
  tcp_init (nids_params.n_tcp_streams);
  ip_frag_init (nids_params.n_hosts);
  scan_init ();
}

void
nids_run ()
{
    if (pcap_loop (desc, -1, (pcap_handler) pcap_hand, 0) <= 0)
	pcap_perror (desc, "loop");
    pcap_close (desc);
}
